import os


class DatabaseManager:
    """
    A database class for easier access
    database MUST be of type sqlite3
    """

    def __str__(self):
        return f"<DATABASE '{self.name}'>"

    def __repr__(self):
        return f"<DATABASE {self.name}, PATH={self.path}, SIZE={self.size}, TABLES={', '.join(self.tables)}>"

    def __with_commit(func):
        def inner(self, *args, **kwargs):
            resp = func(self, *args, **kwargs)
            self.commit()

            return resp

        return inner

    def __with_cursor(func):
        def inner(self, *args, **kwargs):
            cursor = self.database.cursor()
            resp = func(self, cursor, *args, **kwargs)
            cursor.close()

            return resp

        return inner

    def __init__(self, database):
        self.database = database

    def commit(self):
        self.database.commit()

    def close(self):
        self.database.close()

    @property
    def size(self):
        return os.stat(self.path).st_size

    @property
    @__with_cursor
    def tables(self, cursor):
        cursor.execute("SELECT name FROM sqlite_master WHERE type = 'table'")
        return [x[0] for x in cursor.fetchall()]

    @property
    @__with_cursor
    def name(self, cursor):
        cursor.execute("PRAGMA database_list")
        return cursor.fetchall()[0][1]

    @property
    @__with_cursor
    def path(self, cursor):
        cursor.execute("PRAGMA database_list")
        return cursor.fetchall()[0][2]

    def insertifnotexists(self, keys, values, table_name, checks):
        response = self.select([], table_name, checks, True)

        if len(response) == 0:
            self.insert(keys, values, table_name)

    @__with_cursor
    @__with_commit
    def insert(self, cursor, keys, values, table_name):
        query = f"INSERT INTO {table_name} ({', '.join(keys)}) VALUES ({', '.join(['?'] * len(values))})"
        cursor.execute(query, values)

    @__with_cursor
    @__with_commit
    def createtable(self, cursor, table_name, columns, exists=False):
        query = f'CREATE TABLE {"IF NOT EXISTS" if exists else ""} {table_name} ('

        for column in columns:
            query += f"\n{column['name']} {column['type']},"
        query = query[:-1]

        query += "\n)"

        cursor.execute(query)

    @__with_cursor
    @__with_commit
    def update(self, cursor, keys, values, table_name, checks):
        query = f"UPDATE {table_name} SET "

        if keys:
            for key in keys:
                query += f"{key} = ?, "
            query = query[:-2]

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{list(check)[0]} = ? AND "

            query = query[:-4]

        values += [list(x.values())[0] for x in checks]
        cursor.execute(query, values)

    def updateorinsert(self, keys, values, table_name, checks, insertkeys, insertvalues):
        response = self.select([], table_name, checks, True)

        if len(response) == 1:
            self.update(keys, values, table_name, checks)
        else:
            self.insert(insertkeys, insertvalues, table_name)

    @__with_cursor
    @__with_commit
    def delete(self, cursor, table_name, checks):
        query = f"DELETE FROM {table_name} "

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{list(check)[0]} = ? AND "

            query = query[:-4]

        values = [list(x.values())[0] for x in checks]
        cursor.execute(query, values)

    @__with_cursor
    def select(self, cursor, keys, table_name, checks, fetchall=False):
        keys = '*' if not keys else keys
        query = f"SELECT {','.join(keys)} FROM {table_name} "

        if checks:
            query += "WHERE "
            for check in checks:
                query += f"{list(check)[0]} = ? AND "

            query = query[:-4]

        values = [list(x.values())[0] for x in checks]
        cursor.execute(query, values)

        return cursor.fetchall() if fetchall else cursor.fetchone()



import discordSuperUtils
import sqlite3
from discord.ext import commands


database = discordSuperUtils.DatabaseManager(sqlite3.connect("database"))
bot = commands.Bot(command_prefix='-')
LevelingManager = discordSuperUtils.LevelingManager(database, 'xp', bot)


@bot.event
async def on_ready():
    print('Leveling manager is ready.', bot.user)


@LevelingManager.event()
async def on_level_up(message, member_data):
    await message.reply(f"You are now level {member_data['rank']}")


@bot.command()
async def rank(ctx):
    member_data = LevelingManager.get_account(ctx.author)
    await ctx.send(f'You are currently level **{member_data["rank"]}**, with **{member_data["xp"]}** XP.')

bot.run("token")



import discord
import youtube_dl
from .Base import *  # should be .Base, koyashie use Base for testing

# just some options etc.

ytdl_opts = {
    'format': 'bestaudio/best',
    'restrictfilenames': True,
    'noplaylist': False,
    'nocheckcertificate': True,
    'ignoreerrors': False,
    'logtostderr': False,
    'quiet': True,
    'no_warnings': True,
    'default_search': 'auto',
    'source_address': '0.0.0.0'
}

ffmpeg_options = {
    'before_options': '-reconnect 1 -reconnect_streamed 1 -reconnect_delay_max 5', 'options': '-vn'
}

ytdl = youtube_dl.YoutubeDL(ytdl_opts)


# errors/exceptions

class NotPlaying(Exception):
    """Raises error when client is not playing"""


class AlreadyPlaying(Exception):
    """Raises error when player is already playing"""


class NotConnected(Exception):
    """Raises error when client is not connected to a voice channel"""


class NotPaused(Exception):
    """Raises error when player is not paused"""


class QueueEmpty(Exception):
    """Raises error when queue is empty"""


class AlreadyConnected(Exception):
    """Raises error when client is already connected to voice"""


class AlreadyPaused(Exception):
    """Raises error when player is already paused."""


class QueueError(Exception):
    """Raises error when something is wrong with the queue"""


class Player(discord.PCMVolumeTransformer):
    def __init__(self, source, *, data, volume=0.1):
        super().__init__(source, volume)

        self.data = data
        self.title = data.get('title')
        self.url = data.get('url')
        self.duration = data.get('duration')

    def __str__(self):
        return self.title

    def __repr__(self):
        return f'<Player title={self.title}, url={self.url}, duration={self.duration}>'

    @classmethod
    async def make_player(cls, query: str):
        try:
            data = await MusicManager.fetch_data(query)
        except:
            return None

        if 'entries' in data:
            data = data['entries'][0]

        filename = data['url']
        return cls(discord.FFmpegPCMAudio(filename, **ffmpeg_options), data=data)


class QueueManager:
    def __init__(self, volume, queue):
        self.queue = queue
        self.volume = volume
        self.now_playing = None
        self.looping = False

    async def add(self, player):
        self.queue.append(player)

    async def clear(self):
        self.queue.clear()

    async def remove(self, index):
        return self.queue.pop(index)


class MusicManager(EventManager):
    def __init__(self, bot):
        super().__init__()
        self.bot = bot
        self.queue = {}

    def check_queue(self, ctx):
        try:
            if self.queue[ctx.guild.id].looping:
                song = self.queue[ctx.guild.id].queue[0]
                player = Player(discord.FFmpegPCMAudio(song.url, **ffmpeg_options), data=song.data)
                self.queue[ctx.guild.id].queue[0] = player
            else:
                player = self.queue[ctx.guild.id].queue.pop(0)

            self.queue[ctx.guild.id].now_playing = player

            if player is not None and ctx.voice_client:
                player.volume = self.queue[ctx.guild.id].volume
                ctx.voice_client.play(player, after=lambda x: self.check_queue(ctx))  # dont add spaces here after='a'
                if not self.queue[ctx.guild.id].looping:
                    self.bot.loop.create_task(
                        self.call_event('on_play', ctx, player)
                    )

        except IndexError:
            return

    @classmethod
    async def fetch_data(cls, query: str):
        """Returns a dict with info extracted from the URL/query given"""
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(None, lambda: ytdl.extract_info(query, download=False))

    @classmethod
    async def create_player(cls, query):
        return await Player.make_player(query)

    async def queue_add(self, player, ctx):
        """Adds specified player object to queue"""
        if ctx.guild.id in self.queue:
            await self.queue[ctx.guild.id].add(player)
        else:
            self.queue[ctx.guild.id] = QueueManager(0.1, [player])

    async def queue_remove(self, player, ctx):
        """Removed specified player object from queue"""
        if ctx.guild.id in self.queue:
            try:
                self.queue[ctx.guild.id].remove(player)
            except:
                await self.call_event('on_music_error', ctx, QueueError("Failure to remove player from the queue"))

    async def play(self, ctx, player=None):
        """Plays the top of the queue or plays specified player"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        elif player is not None:
            ctx.voice_client.play(player)
            return True

        if not ctx.voice_client.is_playing():
            try:
                player = self.queue[ctx.guild.id].queue[0]
                self.queue[ctx.guild.id].now_playing = player

                if player is not None:
                    ctx.voice_client.play(player, after=lambda x: self.check_queue(ctx))
                    await self.call_event('on_play', ctx, player)
                    return True
            except:
                await self.call_event('on_music_error', ctx, QueueEmpty("Queue is empty."))

    async def pause(self, ctx):
        """Pauses the voice client"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if ctx.voice_client.is_paused():
            await self.call_event('on_music_error', ctx, AlreadyPaused("Player is already paused."))

        ctx.voice_client.pause()
        return True

    async def resume(self, ctx):
        """Resumes the voice client"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if not ctx.voice_client.is_paused():
            await self.call_event('on_music_error', ctx, NotPaused("Player is not paused"))
            return

        ctx.voice_client.resume()
        return True

    async def skip(self, ctx):
        """Most likely wont work"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if not ctx.voice_client.is_playing():
            await self.call_event('on_music_error', ctx, NotPlaying("Client is not playing music"))
            return

        ctx.voice_client.stop()
        return True

    async def volume(self, ctx, volume: int = None):
        """Returns the volume if volume is not given or changes the volume if it is given"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if not ctx.voice_client.is_playing():
            await self.call_event('on_music_error', ctx, NotPlaying("Client is not playing music"))
            return

        if volume is None:
            return ctx.voice_client.source.volume * 100
        else:
            ctx.voice_client.source.volume = volume / 100
            self.queue[ctx.guild.id].volume = volume / 100
            return ctx.voice_client.source.volume * 100

    async def join(self, ctx):
        """Joins voice channel that user is in"""
        if ctx.voice_client and ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, AlreadyConnected("Client is already connected to a voice channel"))
            return

        await ctx.author.voice.channel.connect()
        return True

    async def leave(self, ctx):
        """Leaves voice channel"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        await ctx.voice_client.disconnect()
        return True

    async def now_playing(self, ctx):
        """Returns player of currently playing song"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if not ctx.voice_client.is_playing():
            await self.call_event('on_music_error', ctx, NotPlaying("Client is not playing anything currently"))
            return

        try:
            return self.queue[ctx.guild.id].now_playing
        except:
            await self.call_event('on_music_error', ctx, QueueEmpty("Queue is empty"))

    async def loop(self, ctx):
        """Sets loops to be on or off"""
        if not ctx.voice_client or not ctx.voice_client.is_connected():
            await self.call_event('on_music_error', ctx, NotConnected("Client is not connected to a voice channel"))
            return

        if not ctx.voice_client.is_playing():
            await self.call_event('on_music_error', ctx, NotPlaying("Client is not playing anything currently"))
            return

        try:
            self.queue[ctx.guild.id].looping = not self.queue[ctx.guild.id].looping
            return self.queue[ctx.guild.id].looping
        except IndexError:
            await self.call_event('on_music_error', ctx, QueueEmpty("Queue is empty"))



import asyncio
import discord


class EmojiError(Exception):
    pass


class PageManager:
    def __init__(self, ctx, messages, timeout=60, emojis=None, public=False, index=0):
        self.ctx = ctx
        self.messages = messages
        self.timeout = timeout
        self.emojis = emojis if emojis is not None else ["⏪", "◀️", "▶️", "⏩"]
        self.public = public
        self.index = index

    async def run(self):
        if len(self.emojis) != 4:
            raise EmojiError(f"Passed {len(self.emojis)} emojis when 4 are needed.")

        self.index = 0 if not -1 < self.index < len(self.messages) else self.index

        message_to_send = self.messages[self.index]
        if isinstance(message_to_send, discord.Embed):
            message = await self.ctx.send(embed=message_to_send)
        else:
            message = await self.ctx.send(message_to_send)

        for emoji in self.emojis:
            await message.add_reaction(emoji)

        while True:
            try:
                reaction, user = await self.ctx.bot.wait_for('reaction_add',
                                                             check=lambda x, y: x.message == message,
                                                             timeout=self.timeout)

                if user.bot:
                    continue

                if user != self.ctx.author and not self.public:
                    continue

            except asyncio.TimeoutError:
                break

            if reaction.emoji == self.emojis[0]:
                self.index = 0

            elif reaction.emoji == self.emojis[1]:
                if self.index > 0:
                    self.index -= 1

            elif reaction.emoji == self.emojis[2]:
                if self.index < len(self.messages) - 1:
                    self.index += 1

            elif reaction.emoji == self.emojis[3]:
                self.index = len(self.messages) - 1

            await message.remove_reaction(reaction.emoji, user)
            message_to_send = self.messages[self.index]

            if isinstance(message_to_send, discord.Embed):
                await message.edit(embed=message_to_send, content=None)
            else:
                await message.edit(content=message_to_send, embed=None)


import sqlite3
import discord
import discordSuperUtils
from discord.ext import commands

database = discordSuperUtils.DatabaseManager(sqlite3.connect("database"))
bot = commands.Bot(command_prefix='-')
ReactionManager = discordSuperUtils.ReactionManager(database, 'reaction_roles', bot)


@ReactionManager.event()
async def on_reaction_event(guild, channel, message, member, emoji):
    """This event will be run if there isn't a role to add to the member."""

    if ...:
        print("Created ticket.")


@bot.event
async def on_ready():
    print('Reaction manager is ready.', bot.user)


@bot.command()
async def reaction(ctx, message, emoji: str, remove_on_reaction, role: discord.Role = None):
    message = await ctx.channel.fetch_message(message)

    await ReactionManager.create_reaction(ctx.guild, message, role, emoji, remove_on_reaction)


bot.run("token")



import discordSuperUtils
from discord.ext import commands
import discord


bot = commands.Bot(command_prefix='-')


@bot.event
async def on_ready():
    print('Page manager is ready.', bot.user)


@bot.command()
async def paginator(ctx):
    messages = [
        discord.Embed(
            title='Data (1/2)',
            description="Hello world"
        ),
        "Hello world"
    ]

    await discordSuperUtils.PageManager(ctx, messages).run()


bot.run("token")



from .Leveling import LevelingManager
from .Database import DatabaseManager
from .Music import *
from .Paginator import PageManager
from .ReactionRoles import ReactionManager

__title__ = "discordSuperUtils"
__version__ = "0.0.1"
__author__ = "Koyashie07 & Adam7100"
__license__ = "MIT"


from .Base import EventManager
from .Database import DatabaseManager
from .Paginator import EmojiError

reaction_keys = ['guild', 'message', 'role', 'emoji', 'remove_on_reaction']
reaction_types = ['INTEGER', 'INTEGER', 'INTEGER', 'TEXT', 'INTEGER']


class ReactionManager(EventManager):
    def __init__(self, database: DatabaseManager, table, bot):
        super().__init__()
        self.database = database
        self.table = table
        self.bot = bot

        self.__create_table(reaction_keys, reaction_types, self.table)
        self.bot.add_listener(self.__handle_reactions, "on_raw_reaction_add")
        self.bot.add_listener(self.__handle_reactions, "on_raw_reaction_remove")

    @classmethod
    def format_data(cls, data):
        formatted_data = {}
        for key, value in zip(reaction_keys, data):
            formatted_data[key] = value

        return formatted_data

    @classmethod
    def checks(cls, guild_id, message_id, emoji):
        return [
            {'guild': guild_id},
            {'message': message_id},
            {'emoji': emoji}
        ]

    def __create_table(self, names, types, table_name):
        columns = [{'name': x, 'type': y} for x, y in zip(names, types)]
        self.database.createtable(table_name, columns, True)

    async def __handle_reactions(self, payload):
        if payload.user_id == self.bot.user.id:
            return

        database_checks = self.checks(payload.guild_id,
                                      payload.message_id,
                                      payload.emoji.id if payload.emoji.id is not None else str(payload.emoji))

        reaction_role_data = self.database.select(reaction_keys, self.table, database_checks)

        if not reaction_role_data:
            return

        formatted_data = self.format_data(reaction_role_data)

        guild = self.bot.get_guild(payload.guild_id)
        role = guild.get_role(formatted_data["role"])
        channel = guild.get_channel(payload.channel_id)
        message = await channel.fetch_message(payload.message_id)
        emoji = str(payload.emoji) if payload.emoji.id is None else str(payload.emoji.id)

        if emoji == formatted_data["emoji"]:
            if role is None:
                await self.call_event('on_reaction_event', guild, channel, message, payload.member, emoji)

            else:
                if role not in payload.member.roles:
                    await payload.member.add_roles(role)
                elif formatted_data['remove_on_reaction'] == 1:
                    await payload.member.remove_roles(role)

    async def create_reaction(self, guild, message, role, emoji, remove_on_reaction):
        self.database.insertifnotexists(reaction_keys, [
            guild.id,
            message.id,
            role.id if role is not None else role,
            emoji,
            remove_on_reaction
        ], self.table, self.checks(guild.id, message.id, emoji))

        if len(emoji) > 1:
            emoji = self.bot.get_emoji(emoji)

        try:
            await message.add_reaction(emoji)
        except:
            raise EmojiError("Cannot add reaction to message.")

    async def delete_reaction(self, guild, message, emoji):
        self.database.delete(self.table, self.checks(guild.id, message.id, emoji))


from .Database import DatabaseManager
import time
import math
from .Base import EventManager


database_keys = ['guild', 'member', 'rank', 'xp', 'level_up']


class LevelingManager(EventManager):
    def __init__(self, database: DatabaseManager, table, bot, xp_on_message=5, rank_multiplier=1.5):
        super().__init__()
        self.database = database
        self.table = table
        self.bot = bot
        self.xp_on_message = xp_on_message
        self.rank_multiplier = rank_multiplier

        self.cooldown_members = {}
        self.bot.add_listener(self.__handle_experience, "on_message")
        self.database.createtable(self.table, [{'name': key, 'type': 'INTEGER'} for key in database_keys], True)

    async def __handle_experience(self, message):
        if not message.guild or message.author.bot:
            return

        if message.guild.id not in self.cooldown_members:
            self.cooldown_members[message.guild.id] = {}

        self.create_account(message.author)
        member_timestamp = self.cooldown_members[message.guild.id].get(message.author.id, 0)

        if (time.time() - member_timestamp) >= 10:
            await self.add_experience(message, self.xp_on_message, self.rank_multiplier)
            self.cooldown_members[message.guild.id][message.author.id] = time.time()

    def create_account(self, member):
        self.database.insertifnotexists(database_keys, [member.guild.id, member.id, 1, 0, 50],
                                        self.table, [{'guild': member.guild.id}, {'member': member.id}])

    async def add_experience(self, message, xp, rank_multiplier):
        member_data = self.get_account(message.author)
        if not member_data:
            return

        member_data['xp'] += xp

        leveled_up = False
        while member_data['xp'] >= member_data['level_up']:
            member_data['level_up'] *= rank_multiplier
            member_data['rank'] += 1
            leveled_up = True

        self.update_account(member_data)

        if leveled_up:
            await self.call_event('on_level_up', message, member_data)

    def update_account(self, member_data):
        member_data.pop('next_level_percentage', None)

        self.database.update(list(member_data), list(member_data.values()),
                             self.table, [{'guild': member_data["guild"]}, {'member': member_data["member"]}])

    @classmethod
    def format_data(cls, data):
        formatted_member_data = {}
        for key, value in zip(database_keys + ['next_level_percentage'], data):
            formatted_member_data[key] = value

        return formatted_member_data

    def get_account(self, member):
        member_data = self.database.select([], self.table, [{'guild': member.guild.id}, {'member': member.id}])
        if member_data:
            member_data = self.format_data(member_data)

            initial_rank_xp = 0 if member_data['level_up'] == 50 else member_data['level_up'] / self.rank_multiplier
            percentage = math.floor(abs(member_data['xp'] - initial_rank_xp) /
                                    (member_data['level_up'] - initial_rank_xp) * 100)

            member_data['next_level_percentage'] = percentage
            return member_data
        else:
            return None

    def get_leaderboard(self, guild):
        xp_data = self.database.select([], self.table, [{'guild': guild.id}], True)

        guild_xp_data = [self.format_data(member_data) for member_data in xp_data]

        return sorted(guild_xp_data, key=lambda item: item["xp"], reverse=True)